import cv2


VIDEO_CAPTURE_PROPS_LIST = [
    "CAP_PROP_POS_MSEC",
    "CAP_PROP_POS_FRAMES",
    "CAP_PROP_POS_AVI_RATIO",
    "CAP_PROP_FRAME_WIDTH",
    "CAP_PROP_FRAME_HEIGHT",
    "CAP_PROP_FPS",
    "CAP_PROP_FOURCC",
    "CAP_PROP_FRAME_COUNT",
    "CAP_PROP_FORMAT",
]

VIDEO_CAPTURE_PROPS = {
    prop.split("CAP_PROP_").pop().lower(): prop for prop in VIDEO_CAPTURE_PROPS_LIST
}


def add_props(props):
    def deco(cls):
        for prop in props:
            setattr(cls, prop, VideoCaptureProperty(prop))
        return cls

    return deco


class VideoCaptureProperty:

    _set_err = (
        "The property {p} is not supported by\n"
        "the backend used by the cv2.VideoCapture() instance."
    )
    _docstring = (
        "Alias for cap.get(cv2.{p}) and cap.set(cv2.{p}, value).\n"
        "Raises AttributeError when setting if that property is not supported.\n"
    )

    def __init__(self, name):
        self.name = name
        self.prop = getattr(cv2, VIDEO_CAPTURE_PROPS[name])
        self.__doc__ = self._docstring.format(p=VIDEO_CAPTURE_PROPS[name])

    def __get__(self, obj, objtype=None):
        return obj.cap.get(self.prop)

    def __set__(self, obj, value):
        if not obj.cap.set(self.prop, value):
            raise AttributeError(self._set_err.format(p=VIDEO_CAPTURE_PROPS[self.name]))


@add_props(VIDEO_CAPTURE_PROPS.keys())
class VideoCapture:
    """An adapter for `cv2.VideoCapture`, giving a more Pythonic interface."""

    def __init__(self, *args, **kwargs):
        self.cap = cv2.VideoCapture(*args, **kwargs)
        if not self.cap.isOpened():
            raise ValueError("Unable to open video source:", *args, **kwargs)

    def __getattr__(self, key):
        return getattr(self.cap, key)

    def __iter__(self):
        """Resets the frame position to 0 at the start for repeatable iteration."""
        noread = (False, None)
        if self.cap.isOpened():
            for _, frame in iter(self.cap.read, noread):
                yield frame

    def __enter__(self):
        """Enter the context manager."""
        return self

    def __exit__(self, exctype, exc, exctrace):
        """Releases the video capture object on exiting the context manager."""
        self.cap.release()


class VideoWriter:

    _nowriter = object()

    def __init__(self, filename, fourcc="mp4v", fps=30, frameSize=None, **kwargs):
        self.filename = str(filename)
        self.fourcc = (
            fourcc if isinstance(fourcc, int) else cv2.VideoWriter_fourcc(*fourcc)
        )
        self.fps = fps
        self.kwargs = kwargs

        # wait to create writer based on first frame if size is not provided
        self.writer = (
            self._nowriter if frameSize is None else self._makewriter(frameSize)
        )

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.release()

    def _makewriter(self, frame_size):
        return cv2.VideoWriter(
            filename=self.filename,
            fourcc=self.fourcc,
            fps=self.fps,
            frameSize=frame_size,
            **self.kwargs
        )

    def write(self, frame):
        try:
            return self.writer.write(frame)
        except AttributeError as e:
            if self.writer is self._nowriter:
                h, w = frame.shape[:2]
                self.writer = self._makewriter((w, h))
                self.write(frame)
            else:
                raise e

    def release(self):
        try:
            self.writer.release()
        except AttributeError as e:
            if self.writer is not self._nowriter:
                raise e


class VideoPlayer:

    _actions = {"quit": set(map(ord, "\x1bq"))}

    def __init__(self, cap):
        self.cap = cap
        self.rate = int(1000 / self.cap.fps)

    def play(self, window_name="VideoPlayer", framefunc=None, loop=False):
        """Plays through the video file with OpenCV's imshow()."""

        framefunc = framefunc or (lambda frame: frame)

        for frame in self.cap:
            cv2.imshow(window_name, framefunc(frame))
            key = cv2.waitKey(self.rate) & 0xFF
            if key in self._actions["quit"]:
                return

        if loop:
            return self.play(window_name, framefunc, loop)



import numpy as np
import cv2


def bwdist(
    img,
    method=cv2.DIST_L2,
    dist_mask=cv2.DIST_MASK_5,
    label_type=cv2.DIST_LABEL_CCOMP,
    ravel=True,
):
    """Mimics Matlab's bwdist function, similar to OpenCV's distanceTransform()
    but with different output.

    https://www.mathworks.com/help/images/ref/bwdist.html

    Available metrics:
        https://docs.opencv.org/3.4/d7/d1b/group__imgproc__misc.html#gaa2bfbebbc5c320526897996aafa1d8eb

    Available distance masks:
        https://docs.opencv.org/3.4/d7/d1b/group__imgproc__misc.html#gaaa68392323ccf7fad87570e41259b497

    Available label types:
        https://docs.opencv.org/3.4/d7/d1b/group__imgproc__misc.html#ga3fe343d63844c40318ee627bd1c1c42f
    """
    flip = cv2.threshold(img, 0, 255, cv2.THRESH_BINARY_INV)[1]
    dist, labeled = cv2.distanceTransformWithLabels(flip, method, dist_mask)

    if ravel:  # return linear indices if ravel == True (default)
        idx = np.zeros(img.shape, dtype=np.intp)
        idx_func = np.flatnonzero
    else:  # return two-channel indices if ravel == False
        idx = np.zeros((*img.shape, 2), dtype=np.intp)
        idx_func = lambda masked: np.dstack(np.where(masked))

    for l in np.unique(labeled):
        mask = labeled == l
        idx[mask] = idx_func(img * mask)
    return dist, idx


def imfill(bin_img):
    """Fills holes in the input binary image.

    Achieves the same output as the imfill(BW, 'holes') variant.

    https://www.mathworks.com/help/images/ref/imfill.html
    """
    contours = cv2.findContours(bin_img, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)[1]
    return cv2.drawContours(bin_img, contours, -1, 255, -1)



"""Utilities for computer vision."""

from .clustering import kmeans_periodic
from .display import *
from . import hough
from .imgproc import circular_mask, circular_gradient, resize_pad
from .iterators import LineIterator
from . import matlab
from .transformations import warpAffinePadded, warpPerspectivePadded
from .videoio import VideoCapture, VideoPlayer, VideoWriter
from .imgio import *
from .types import *

__version__ = "0.4.4"



import cv2
import numpy as np


TWO_PI = 2 * np.pi


def kmeans_periodic(columns, intervals, data, *args, **kwargs):
    """Runs kmeans with periodicity in a subset of dimensions.
    
    Transforms columns with periodicity on the specified intervals into two
    columns with coordinates on the unit circle for kmeans. After running
    through kmeans, the centers are transformed back to the range specified
    by the intervals.

    Arguments
    ---------
    columns : sequence
        Sequence of indexes specifying the columns that have periodic data
    intervals : sequence of length-2 sequences
        Sequence of (min, max) intervals, one interval per column
    See help(cv2.kmeans) for all other arguments, which are passed through.

    Returns
    -------
    See help(cv2.kmeans) for outputs, which are passed through; except centers,
    which is modified so that it returns centers corresponding to the input
    data, instead of the transformed data.
    
    Raises
    ------
    cv2.error
        If len(columns) != len(intervals)
    """

    # Check each periodic column has an associated interval
    if len(columns) != len(intervals):
        raise cv2.error("number of intervals must be equal to number of columns")

    ndims = data.shape[1]
    ys = []

    # transform each periodic column into two columns with the x and y coordinate
    # of the angles for kmeans; x coord at original column, ys are appended
    for col, interval in zip(columns, intervals):
        a, b = min(interval), max(interval)
        width = b - a
        data[:, col] = TWO_PI * (data[:, col] - a) / width % TWO_PI
        ys.append(width * np.sin(data[:, col]))
        data[:, col] = width * np.cos(data[:, col])

    # append the ys to the end
    ys = np.array(ys).transpose()
    data = np.hstack((data, ys)).astype(np.float32)

    # run kmeans
    retval, bestLabels, centers = cv2.kmeans(data, *args, **kwargs)

    # transform the centers back to range they came from
    for i, (col, interval) in enumerate(zip(columns, intervals)):
        a, b = min(interval), max(interval)
        angles = np.arctan2(centers[:, ndims + i], centers[:, col]) % TWO_PI
        centers[:, col] = a + (b - a) * angles / TWO_PI
    centers = centers[:, :ndims]

    return retval, bestLabels, centers



import cv2
import numpy as np
from collections import defaultdict
from bisect import bisect


# Constants


RHOTHETA = "rhotheta"
ENDPOINT = "endpoint"
ERRORVAL = np.nan


# Utility Functions


def linetype(line):
    """Returns 'rhotheta' if line is defined by two variables,
    i.e. [[rho, theta]], or 'endpoint' if line is defined by four
    variables, i.e. [[x1, y1, x2, y2]].

    Parameters
    ----------
    line : np.ndarray
        Either np.array([[rho, theta]]) or np.array([[x1, y1, x2, y2]])
        where theta is in radians.

    Returns
    -------
    linetype : str
        'rhotheta' if line contains two vars; 'endpoint' if four.

    Example
    -------
    >>> import numpy as np
    >>> rho, theta = 5, np.pi/4
    >>> line_rhotheta = np.array([[rho, theta]])
    >>> x1, y1, x2, y2 = 0, 0, 50, 100
    >>> line_endpoint = np.array([[x1, y1, x2, y2]])
    >>> linetype(line_rhotheta)
    'rhotheta'
    >>> linetype(line_endpoint)
    'endpoint'
    """

    if len(line[0]) == 2:
        return RHOTHETA
    elif len(line[0]) == 4:
        return ENDPOINT
    else:
        raise TypeError


def lineangle(line):
    """Returns the angle (in radians) of the line in [0, 2π).

    Lines in rho-theta form are only defined with theta in [0, π)
    but allows negative rho values; this function takes into account
    the sign of rho to determine the angle in [0, 2π).

    Similarly, lines in endpoint form will return different angles if
    their endpoints are flipped.

    Parameters
    ----------
    line : np.ndarray
        Either np.array([[rho-theta]]) or np.array([[x1, y1, x2, y2]])
        where `theta` is in radians.

    Returns
    -------
    angle : float
        Angle in radians in the interval [0, 2π) of the line.

    Example
    -------
    >>> import numpy as np
    >>> rho, theta = -5, 3*np.pi/4
    >>> line_rhotheta = create_line(rho, theta)
    >>> x1, y1, x2, y2 = 0, 0, 50, 50
    >>> line_endpoint = create_line(x1, y1, x2, y2)
    >>> print('lineangle(line_rhotheta):', lineangle(line_rhotheta))
    lineangle(line_rhotheta): 0.785398163397
    >>> print('lineangle(line_endpoint):', lineangle(line_endpoint))
    lineangle(line_endpoint): 0.785398163397
    >>> print('radian(45 degrees):', np.pi/4)
    radian(45 degrees): 0.7853981633974483
    """

    _linetype = linetype(line)
    if _linetype == RHOTHETA:
        # actual angle of line is perpendicular to theta
        rho, theta = line[0]
        if rho >= 0:
            angle = theta + np.pi / 2
        else:
            angle = theta - np.pi / 2
        return angle % (2 * np.pi)
    elif _linetype == ENDPOINT:
        x1, y1, x2, y2 = line[0]
        angle = np.arctan2(y2 - y1, x2 - x1)
        return angle % (2 * np.pi)
    else:
        raise TypeError


def isparallel(line1, line2, tol=None):
    """Boolean to check if two lines are parallel.

    Parameters
    ----------
    line1 : np.ndarray
        Either rho-theta or endpoint form.
    line2 : np.ndarray
        Either rho-theta or endpoint form; does not need to be the same
        as line1.
    tol : float
        An optional tolerance value for parallelism check. If tol=0,
        the function checks for exact parallelism (that is, the lines
        have the exact same angle to machine precision). By default,
        the tolerance is a single arc-second: tol=np.pi/(180*60*60).

    Returns
    -------
    isparallel : bool
        True if the difference in angle between each line is less than
        tol, False otherwise. 

    Example
    -------
    >>> line1 = create_line(25, np.pi/4)
    >>> line2 = create_line(3, np.pi/4)
    >>> isparallel(line1, line2)
    True
    """
    tol = np.pi / (180 * 60 * 60) if tol == None else tol

    angle1 = lineangle(line1) % np.pi
    angle2 = lineangle(line2) % np.pi
    diff = abs(angle1 - angle2)
    diff = min(diff, np.pi - diff)
    return diff < tol


def length(line):
    """Get the length of a 2-D line segment using Euclidean distance.

    Parameters
    ----------
    line : np.ndarray
        A line in endpoint form.
    
    Returns
    -------
    length : float
        Euclidean length of the line.

    Example
    -------
    >>> line = create_line(0, 0, 100, 100)
    >>> length(line)
    141.42135623730951
    """
    x1, y1, x2, y2 = line[0]
    return np.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)


def create_line_rhotheta(rho, theta):
    """Create a line in rho-theta form.

    Parameters
    ----------
    rho : int
        The radius of the circle centered at (0, 0) which the line is
        tangent to.
    theta : float
        The angle (in radians) about the origin; the line is defined
        perpendicularly to this angle.

    Returns
    -------
    line : np.ndarray
        A line in the form np.array([[rho, theta]]). The angle returned
        will be in [0, π), with negative rho values for angles in the
        interval [π, 2π).

    Example
    -------
    >>> import numpy as np
    >>> rho, theta = 5, np.pi/4
    >>> create_line_rhotheta(rho, theta)
    array([[ 5.        ,  0.78539816]])
    >>> create_line(3, 5*np.pi/4)
    array([[-3.        ,  0.78539816]])
    """
    theta = theta % (2 * np.pi)
    if theta < np.pi:
        return np.array([[rho, theta]])
    return np.array([[-rho, theta % np.pi]])


def create_line_endpoint(x1, y1, x2, y2):
    """Create a line in endpoint form.

    Parameters
    ----------
    x1, y1, x2, y2 : numeric
        The endpoints of the line are (x1, y1) and (x2, y2). Points
        can be specified with ints or float subpixel points.

    Returns
    -------
    line : np.ndarray
        A line in the form np.array([[x1, y1, x2, y2]]).

    Example
    -------
    >>> x1, y1, x2, y2 = 0, 0, 500, 500
    >>> create_line_endpoint(x1, y1, x2, y2)
    array([[  0,   0, 500, 500]])
    """
    return np.array([[x1, y1, x2, y2]])


def create_line(*args):
    """Create a line in rho-theta or endpoint form, depending on the
    number of arguments specified.

    Parameters
    ----------
    *args : numeric
        Either two numbers specifying `rho, theta` or four numbers
        specifying `x1, y1, x2, y2`. 

    Returns
    -------
    line : np.ndarray
        A line in rho-theta form `np.array([[rho, theta]])` or endpoint
        form `np.array([[x1, y1, x2, y2]])`. If it is in rho-theta form,
        the angle returned will be in `[0, π)`, with negative `rho` values
        for angles input in `[π, 2π)`.

    Example
    -------
    >>> import numpy as np
    >>> rho, theta = 5, np.pi/4
    >>> create_line(rho, theta)
    array([[ 5.        ,  0.78539816]])
    >>> x1, y1, x2, y2 = 0, 0, 500, 500
    >>> create_line(x1, y1, x2, y2)
    array([[  0,   0, 500, 500]])
    
    """
    if len(args) == 2:
        return create_line_rhotheta(*args)
    elif len(args) == 4:
        return create_line_endpoint(*args)
    else:
        raise TypeError


# Distance Functions


def point_line_dist(point, line):
    """See
    https://en.wikipedia.org/wiki/Distance_from_a_point_to_a_line#Line_defined_by_two_points
    """

    _linetype = linetype(line)
    if linetype(line) == RHOTHETA:
        line = convert(line)
    for x1, y1, x2, y2 in line:
        for x0, y0 in point:
            numerator = abs((y2 - y1) * x0 - (x2 - x1) * y0 + x2 * y1 - y2 * x1)
            denominator = length(line)
            if denominator < 1:
                print("Error: line length less than a single pixel.")
                return ERRORVAL
    return numerator / denominator


def point_point_dist(point1, point2):
    return length(np.array([[*point1, *point2]]))


# Conversion Functions


def endpoint(line, bbox=[0, 0, 1e5, 1e5]):
    """Converts a line from rho-theta to endpoint form.

    Endpoints are given by the intersection to the input bounding box.

    Parameters
    ----------
    line : np.ndarray
        A line in rho-theta form.
    bbox : list, optional
        A bounding box defined by the top-left and bottom-right points
        or in an alternative view by the left-top-right-bottom walls,
        in the form [x1, y1, x2, y2]; default [0, 0, 1e5, 1e5].
        Intersections are found between the input line and the border
        lines of the bounding box. Only the two points touching the 
        border of the bounding box are returned. For a specific image
        height and width, set bbox=[0, 0, w, h].
    
    Returns
    -------
    line : np.ndarray
        Returns an endpoint line with endpoints at the intersection
        with bounding box borders.
    """
    x1, y1, x2, y2 = bbox
    border_lines = [[[x1, np.pi / 2]], [[x2, np.pi / 2]], [[y1, 0]], [[y2, 0]]]

    interx = [intersection(line, border_lines[i]) for i in range(4)]

    bounded_interx = []
    for point in interx:
        for x, y in point:
            if (x1 <= x <= x2) and (y1 <= y <= y2):
                bounded_interx.extend([x, y])
        if len(bounded_interx) == 4:
            # in case a line hits the corner exactly
            break
    if len(bounded_interx) != 4:
        # line is outside the bounding box
        return [[[ERRORVAL, ERRORVAL]]] * 4
    return np.array([bounded_interx])


def rhotheta(line):
    """Converts a line from endpoint to rho-theta form.

    Infinitely extending rho-theta lines are returned.

    Parameters
    ----------
    line : np.ndarray
        A line in endpoint form.
    
    Returns
    -------
    line : np.ndarray
        An infinitely extending rho-theta line.
    """
    theta = lineangle(line)
    sign = 1 if theta < np.pi else -1
    theta = theta % np.pi
    rho = sign * point_line_dist([[0, 0]], line)
    return np.array([[rho, theta]])


def convert(line, bbox=[0, 0, 1e5, 1e5]):
    """Converts a line between rho-theta and endpoint form.

    When rho-theta lines are input, endpoints are given by the
    intersection to the input bounding box. When endpoint lines
    are input, infinitely extending rho-theta lines are returned.

    Parameters
    ----------
    line : np.ndarray or list
        A line in rho-theta or endpoint form, to be converted
        to the other; *or* a list (possibly nested) of such lines.
    bbox : list, optional
        Used only when converting from rho-theta to endpoint lines. 
        A bounding box defined by the top-left and bottom-right points
        or in an alternative view by the left-top-right-bottom walls,
        in the form [x1, y1, x2, y2]; default [0, 0, 1e5, 1e5].
        Intersections are found between the input line and the border
        lines of the bounding box. Only the two points touching the 
        border of the bounding box are returned. For a specific image
        height and width, set bbox=[0, 0, w, h].
    
    Returns
    -------
    line : np.ndarray
        If the input is a rho-theta line, returns an endpoint line
        with intersections at the bounding box borders. If the input
        line is an endpoint line, returns a rho-theta line. If the 
        input is a list of lines, then returns the converted list.
    """
    if isinstance(line, list):
        return [convert(l, bbox) for l in line]
    _linetype = linetype(line)
    if _linetype == RHOTHETA:
        return endpoint(line, bbox)
    elif _linetype == ENDPOINT:
        return rhotheta(line)
    else:
        raise TypeError


def pointslope(line, x=0, tol=None):
    """Convert from rho-theta or endpoint line to point-slope form.

    Parameters
    ----------
    line : np.ndarray
        A rho-theta or endpoint line.
    x : numeric, optional
        The point on the line returned will pass through the vertical
        line at x. By default, x=0, rendering this function identical
        except in output format to slopeintercept(line).
    tol : float, optional
        Tolerance for checking verticalism. See docs of isparallel().

    Returns
    -------
    line : np.ndarray
        A line in the form np.array([[x, y, m]]) where (x, y) is a
        point on the line and m is the slope of the line. In the case
        of a vertical line, [[x', np.nan, np.nan]] will be returned,
        where x' is the location of the line. In any other case, the
        returned line will pass through the vertical line defined by x.
    """
    _linetype = linetype(line)

    if _linetype == RHOTHETA:
        rho, theta = line[0]
        vline = create_line(x, 0)
        if isparallel(line, vline):
            return np.array([[rho, ERRORVAL, ERRORVAL]])
        m = np.sign(rho) * np.tan(theta + np.pi / 2)
    elif _linetype == ENDPOINT:
        x1, y1, x2, y2 = line[0]
        if isparallel(line, vline):
            return np.array([[x1, ERRORVAL, ERRORVAL]])
        m = (y2 - y1) / (x2 - x1)
        vline = create_line(x, 0, x, 1)

    interx = intersection(line, vline)
    x, y = interx[0]

    return np.array([[x, y, m]])


def slopeintercept(line, tol=None):
    """Convert from rho-theta or endpoint line to slope-intercept form.

    Parameters
    ----------
    line : np.ndarray
        A rho-theta or endpoint line.
    tol : float, optional
        Tolerance for checking verticalism. See docs of isparallel().

    Returns
    -------
    line : np.ndarray
        A line in the form np.array([[m, b]]) where m is the slope of
        the line and b is the y-intercept. In the case of a vertical
        line, [[np.nan, np.nan]] will be returned; in any other case,
        the returned line will follow the Cartesian equation y=mx+b.
    """
    x, y, m = pointslope(line, tol)
    if x == 0:
        return np.array([[m, y]])
    else:
        return np.array([[ERRORVAL, ERRORVAL]])


# Segmenting Functions


def segment_angle_kmeans(lines, k=2, **kwargs):
    """Groups lines based on angle with k-means.

    Computes the angle of each line and uses k-means on the coordinates
    of the angle on the unit circle to segment k angles inside lines.
    """

    # Define criteria = (type, max_iter, epsilon)
    default_criteria_type = cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER
    criteria = kwargs.get("criteria", (default_criteria_type, 10, 1.0))
    flags = kwargs.get("flags", cv2.KMEANS_RANDOM_CENTERS)
    attempts = kwargs.get("attempts", 10)

    # returns angles in [0, π)
    angles = [lineangle(line) % np.pi for line in lines]
    # stretch angles to the full range and map to coords
    pts = np.array(
        [[[np.cos(2 * angle), np.sin(2 * angle)]] for angle in angles], dtype=np.float32
    )
    # run kmeans on unit coordinates of the angle
    labels, centers = cv2.kmeans(pts, k, None, criteria, attempts, flags)[1:]
    labels = labels.reshape(-1)  # transpose to row vec

    # segment lines based on their kmeans label
    segmented = defaultdict(list)
    for i, line in zip(range(len(lines)), lines):
        segmented[labels[i]].append(line)
    segmented = list(segmented.values())
    return segmented


def segment_angle_range(lines, step=np.pi / 2):
    """Groups lines based on angle.

    Chops up the interval [0, π) into ceil(π/step) bins and assigns
    each line to a bin. The first bin is only half-width and merged
    with the last bin so lines that are nearly horizontal are counted
    in the same bin.
    """
    angle_cuts = np.arange(step / 2, np.pi, step=step)
    segmented = defaultdict(list)
    for line in lines:
        angle_bin = bisect(angle_cuts, lineangle(line) % np.pi) % len(angle_cuts)
        segmented[angle_bin].append(line)
    segmented = list(segmented.values())
    return segmented


def segment_angle_linspace(lines, num=2):
    """Groups lines based on angle.

    Chops up the interval [0, π) into num bins and assigns each line
    to a bin. The first bin is only half-width and merged with the last
    bin so lines that are nearly horizontal are counted in the same bin.
    """
    step = np.pi / num
    return segment_angle_range(lines, step)


# Intersection Functions


def rhotheta_intersection(line1, line2, tolerance=1e-6, subpixel=False):
    """Finds the intersection of two lines in rho, theta form.

    Returns closest integer pixel locations. Returns nan if lines are
    parallel or very close to parallel.

    See https://stackoverflow.com/a/383527/5087436
    """
    rho1, theta1 = line1[0]
    rho2, theta2 = line2[0]

    parallelism = abs(theta1 - theta2)
    if parallelism < tolerance:
        return [[ERRORVAL, ERRORVAL]]

    # Ax = b    linear system
    A = np.array([[np.cos(theta1), np.sin(theta1)], [np.cos(theta2), np.sin(theta2)]])
    b = np.array([[rho1], [rho2]])
    x0, y0 = np.linalg.solve(A, b)

    if subpixel:
        return [[x0, y0]]
    x0, y0 = int(np.round(x0)), int(np.round(y0))
    return [[x0, y0]]


def endpoint_intersection(line1, line2, tolerance=1e-6, subpixel=False):
    """Finds the intersection of two lines in endpoint form.

    Returns closest integer pixel locations. Returns nan if lines are
    parallel or very close to parallel.

    See https://stackoverflow.com/a/383527/5087436
    """

    parallelism = abs(_line_angle(line1) - _line_angle(line2))
    if parallelism < tolerance:
        return [[ERRORVAL, ERRORVAL]]

    # extract points
    x1, y1, x2, y2 = line1[0]
    x3, y3, x4, y4 = line2[0]

    # compute determinant
    denominator = (x1 - x2) * (y3 - y4) - (y1 - y2) * (x3 - x4)
    x0 = (
        (x1 * y2 - y1 * x2) * (x3 - x4) - (x1 - x2) * (x3 * y4 - y3 * x4)
    ) / denominator
    y0 = (
        (x1 * y2 - y1 * x2) * (y3 - y4) - (y1 - y2) * (x3 * y4 - y3 * x4)
    ) / denominator

    if subpixel:
        return [[x0, y0]]
    x0, y0 = int(np.round(x0)), int(np.round(y0))
    return [[x0, y0]]


def intersection(line1, line2, tolerance=1e-6, subpixel=False):
    """Finds the intersection of two lines.

    Returns closest integer pixel locations. Returns np.nan if lines are
    parallel or very close to parallel, or if two lines are not defined
    in rho, theta or endpoint form.
    """
    linetype1 = linetype(line1)
    linetype2 = linetype(line2)

    if linetype1 == linetype2 == RHOTHETA:
        return rhotheta_intersection(line1, line2, tolerance, subpixel)
    elif linetype1 == linetype2 == ENDPOINT:
        return endpoint_intersection(line1, line2, tolerance, subpixel)
    else:
        print("Line type error; nan intersection returned")
        return [[ERRORVAL, ERRORVAL]]


# Drawing Functions


def draw_lines(img, lines, color=None, thickness=1):
    h, w = img.shape[:2]
    color = color or (0, 255, 0) if len(img.shape) == 3 else 255
    _linetype = linetype(lines[0])
    if _linetype == RHOTHETA:
        lines = [convert(line, bbox=[0, 0, w, h]) for line in lines]
    for line in lines:
        for x1, y1, x2, y2 in line:
            if ERRORVAL in [x1, y1, x2, y2]:
                print("Line {} has np.nan vals; not drawing.".format(line))
                continue
            cv2.line(img, (x1, y1), (x2, y2), color, thickness)
    return img


def mark_endpoints(img, lines, color=None, markerType=cv2.MARKER_CROSS, markerSize=5):
    color = color or (0, 255, 255) if len(img.shape) == 3 else 255
    for line in lines:
        for x1, y1, x2, y2 in line:
            if ERRORVAL in [x1, y1, x2, y2]:
                print("Line {} has np.nan vals; not drawing.".format(line))
                continue
            cv2.drawMarker(img, (x1, y1), color, markerType, markerSize)
            cv2.drawMarker(img, (x2, y2), color, markerType, markerSize)
    return img


def mark_intersections(
    img, intersections, color=None, markerType=cv2.MARKER_TRIANGLE_UP, markerSize=5
):
    color = color or (0, 255, 255) if len(img.shape) == 3 else 255
    for point in intersections:
        for x, y in point:
            if ERRORVAL in [x, y]:
                print("Point {} has np.nan vals; not drawing.".format(point))
                continue
            cv2.drawMarker(img, (x, y), color, markerType, markerSize)
    return img



import cv2
from pathlib import Path


def imread(imgpath, *args, **kwargs):
    """Reads an image, providing helpful errors on failed reads.

    Allows pathlib.Path objects as well as strings for the imgpath.
    """
    img = cv2.imread(str(imgpath), *args, **kwargs)
    if img is None:
        p = Path(imgpath)
        if p.exists():
            raise ValueError("Image is empty!")
        raise FileNotFoundError(f"Image path {p.absolute()} doesn't exist!")

    return img


def imwrite(imgpath, img, *args, **kwargs):
    """Writes an image, providing helpful errors on failed writes.

    Allows pathlib.Path objects as well as strings for the imgpath.
    Will create the directories included in the imgpath if they don't exist.
    """
    if img is None:
        raise ValueError("Image is empty!")
    Path(imgpath).parent.mkdir(parents=True, exist_ok=True)
    imgpath = str(imgpath)
    return cv2.imwrite(imgpath, img, *args, **kwargs)



import cv2


def imshow(img, wait=0, window_name=""):
    if img is None:
        raise ValueError(
            "Image is empty; ensure you are reading from the correct path."
        )
    cv2.imshow(window_name, img)
    return cv2.waitKey(wait) & 0xFF


def imshow_ipython(img):
    """Shows an image in a Jupyter notebook.
    Raises ValueError if img is None or if img cannot be encoded.
    """
    if img is None:
        raise ValueError("Image has no data (img is None).")

    success, encoded = cv2.imencode(".png", img)
    if not success:
        raise ValueError("Error encoding image.")

    from IPython.display import Image, display

    display(Image(encoded))


def imshow_components(labels, *args, **kwargs):
    # Map component labels to hue val
    label_hue = np.uint8(179 * labels / np.max(labels))
    blank_ch = 255 * np.ones_like(label_hue)
    labeled_img = cv2.merge([label_hue, blank_ch, blank_ch])

    # cvt to BGR for display
    labeled_img = cv2.cvtColor(labeled_img, cv2.COLOR_HSV2BGR)

    # set bg label to black
    labeled_img[label_hue == 0] = 0
    return imshow(labeled_img, *args, **kwargs)


def imshow_autoscale(img, *args, **kwargs):
    scaled = cv2.normalize(img, None, 0, 255, cv2.NORM_MINMAX, cv2.CV_8U)
    return imshow(scaled, *args, **kwargs)


def imshow_enlarged(img, scale=10, grid=True, color=200, wait=0, window_name=""):
    h, w = img.shape[:2]
    r = cv2.resize(img, None, fx=scale, fy=scale, interpolation=cv2.INTER_NEAREST)
    if grid:
        r = add_grid(img, scale, color)
    return imshow(r, wait, window_name)


def add_grid(img, scale=10, color=200):

    h, w = img.shape[:2]
    r = cv2.resize(img, None, fx=scale, fy=scale, interpolation=cv2.INTER_NEAREST)

    partial_grid = np.zeros((scale, scale), dtype=bool)
    partial_grid[0, :] = True
    partial_grid[:, 0] = True
    gridlines = np.tile(partial_grid, (h, w))
    r[gridlines] = color

    pad_sizes = ((0, 1), (0, 1), (0, 0)) if len(img.shape) == 3 else ((0, 1), (0, 1))
    r = np.pad(r, pad_sizes, "constant", constant_values)
    r[-1, :] = color
    r[:, -1] = color

    return r


# Display image in browser (for working in interpreters without a GUI)


from http.server import BaseHTTPRequestHandler, HTTPServer
import webbrowser
from concurrent.futures import ThreadPoolExecutor
import base64


def _html_imshow(img):
    success, encoded_img = cv2.imencode(".png", img)

    html = """
<html>
<title>cvtools/imshow</title>
<head> 
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1">
</head>
<body style="background-color:#16151a;margin:0;">
<div style="display:flex;justify-content:center;align-items:center;height:100vh;">
<img src="data:image/png;base64,{encoded_img}"/>
</div>
</body>
</html>
"""

    b64_encoded_img = base64.b64encode(encoded_img).decode() if success else ""
    return html.format(encoded_img=b64_encoded_img)


class _ImshowRequestHandler(BaseHTTPRequestHandler):

    # handle GET request from browser
    def do_GET(self):
        self.send_response(200)
        if self.path == f"/{self.imshow_route}":
            self.send_header("Content-type", "text/html")
            self.end_headers()
            self.wfile.write(_html_imshow(self.imshow_img).encode())
        return

    # remove logging statements by returning nothing here
    def log_message(self, format, *args):
        return


def imshow_browser(img, server="localhost", port=32830, route="imshow"):
    """Display an image in a browser. 
    
    Spins up a server to serve the image for a single request.
    Opens the browser to make that request, then shuts down the server.
    """

    class ImshowRequestHandler(_ImshowRequestHandler):
        imshow_img = img
        imshow_route = route

    host = "localhost"
    port = 32830
    server = HTTPServer((host, port), ImshowRequestHandler)
    with ThreadPoolExecutor(max_workers=1) as executor:
        # handle_request() blocks, so submit in an executor.
        # the browser can open the window and get served the image,
        # at which point the submitted task is completed.
        future = executor.submit(server.handle_request)
        webbrowser.open_new(f"http://{host}:{port}/{route}")



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['cvtools']

package_data = \
{'': ['*']}

setup(name='cvtools-alkasm',
      version='0.4.4',
      description='Utilities for computer vision.',
      author='Alexander Reynolds',
      author_email='ar@reynoldsalexander.com',
      url='https://github.com/alkasm/cvtools',
      packages=packages,
      package_data=package_data,
      python_requires='>=3.6',
     )



"""
This module mimics some of the OpenCV built-in types that don't get translated
into Python directly. Generally, these just get mapped to/from tuples.

However, a lot of OpenCV code would benefit from types that describe the tuple,
in addition to named attributes. As OpenCV expects tuples for these datatypes,
subclassing from tuples and namedtuples allows for flexibility without breaking
capability.

Wherever it made sense, functionality was copied from OpenCV. For overloaded
CPP functions, there is not a hard rule for deciding which version becomes the
defacto Python method, but in some cases, both have been provided. Alternative
constructors are given in the usual Python way---as classmethods prepended with
the verb "from".

Some liberties have been taken with respect to naming. In particular, camelCase
method names have been translated to snake_case. Some methods are provided that
OpenCV doesn't contain. These methods are prepended with an underscore to show
that they are not part of OpenCV's API, but they are intended for use.

Aside from Scalars (where you can use a numpy array for vector operations),
the usual arithmetic operations available in OpenCV are available here.

You can average two points by adding them and dividing by two; you can add a
point to a rect to shift it; and so on.
"""

from typing import NamedTuple
import numpy as np
import cv2 as cv


class _IterOps:
    """Generic mixin for arithmetic operations shared between Point & Size classes."""

    def __add__(self, other):
        return type(self)(a + b for a, b in zip(self, other))

    def __sub__(self, other):
        return type(self)(a - b for a, b in zip(self, other))

    def __mul__(self, other):
        if hasattr(other, "__iter__"):
            return type(self)(a * b for a, b in zip(self, other))
        return type(self)(a * other for a in self)  # scalar multiplication

    def __div__(self, other):
        if hasattr(other, "__iter__"):
            return type(self)(a / b for a, b in zip(self, other))
        return type(self)(a / other for a in self)  # scalar division

    def __eq__(self, other):
        return all(a == b for a, b in zip(self, other))


class Point(NamedTuple, _IterOps):
    x: float
    y: float

    def cross(self, point):
        return float(np.cross(self, point))

    def dot(self, point):
        return float(np.dot(self, point))

    def ddot(self, point):
        return self.dot(point)

    def inside(self, rect):
        """checks whether the point is inside the specified retangle"""
        rect = Rect(*rect)
        return rect.contains(self)


class Point3(NamedTuple, _IterOps):
    x: float
    y: float
    z: float

    def cross(self, point):
        return type(self)(*np.cross(self, point))

    def dot(self, point):
        return float(np.dot(self, point))

    def ddot(self, point):
        return self.dot(point)


class Size(NamedTuple, _IterOps):
    width: float
    height: float

    def area(self):
        return self.height * self.width

    def empty(self):
        """true if empty"""
        return self.width <= 0 or self.height <= 0

    @classmethod
    def from_image(cls, image):
        h, w = image.shape[:2]
        return cls(w, h)


class Rect(NamedTuple):
    x: float
    y: float
    width: float
    height: float

    def tl(self):
        """top left point"""
        return Size(self.x, self.y)

    def br(self):
        """bottom right point"""
        return Size(self.x + self.width, self.y + self.height)

    def area(self):
        return self.height * self.width

    def size(self):
        """size (width, height) of the rectangle"""
        return Size(self.width, self.height)

    def contains(self, point):
        """checks whether the rectangle contains the point"""
        point = Point(*point)
        return (
            self.x <= point.x <= self.x + self.width
            and self.y <= point.y <= self.y + self.height
        )

    def empty(self):
        """true if empty"""
        return self.width <= 0 or self.height <= 0

    def __add__(self, other):
        """Shift or alter the size of the rectangle.
        𝚛𝚎𝚌𝚝 ± 𝚙𝚘𝚒𝚗𝚝 (shifting a rectangle by a certain offset)
        𝚛𝚎𝚌𝚝 ± 𝚜𝚒𝚣𝚎 (expanding or shrinking a rectangle by a certain amount)
        """
        if isinstance(other, Point):
            origin = Point(self.x + other.x, self.y + other.y)
            return self.from_origin(origin, self.size)
        elif isinstance(other, Size):
            size = Size(self.width + other.width, self.height + other.height)
            return self.from_origin(self.tl(), size)
        raise TypeError(
            "Adding to a rectangle generically is ambiguous.\n"
            "Add a Point to shift the top-left point, or a Size to expand the rectangle."
        )

    def __sub__(self, other):
        """Shift or alter the size of the rectangle.
        𝚛𝚎𝚌𝚝 ± 𝚙𝚘𝚒𝚗𝚝 (shifting a rectangle by a certain offset)
        𝚛𝚎𝚌𝚝 ± 𝚜𝚒𝚣𝚎 (expanding or shrinking a rectangle by a certain amount)
        """
        if isinstance(other, Point):
            origin = Point(self.x - other.x, self.y - other.y)
            return self.from_origin(origin, self.size)
        elif isinstance(other, Size):
            size = Size(self.width - other.width, self.height - other.height)
            return self.from_origin(self.tl(), size)
        raise TypeError(
            "Subtracting from a rectangle generically is ambiguous.\n"
            "Subtract a Point to shift the top-left point, or a Size to shrink the rectangle."
        )

    def __and__(self, other):
        """rectangle intersection"""
        other = type(self)(*other)
        x = max(self.x, other.x)
        y = max(self.y, other.y)
        w = min(self.x + self.width, other.x + other.width) - x
        h = min(self.y + self.height, other.y + other.height) - y

        return type(self)(0, 0, 0, 0) if (w <= 0 or h <= 0) else type(self)(x, y, w, h)

    def __or__(self, other):
        """minimum area rectangle containing self and other."""
        other = type(self)(*other)
        if self.empty():
            return other
        elif not other.empty():
            x = min(self.x, other.x)
            y = min(self.y, other.y)
            w = max(self.x + self.width, other.x + other.width) - x
            h = max(self.y + self.height, other.y + other.height) - y
            return type(self)(x, y, w, h)
        return type(self)(0, 0, 0, 0)

    def __eq__(self, other):
        other = type(self)(*other)
        return all(a == b for a, b in zip(self, other))

    @classmethod
    def from_points(cls, top_left, bottom_right):
        """Alternative constructor using two points."""
        x1, y1 = top_left
        x2, y2 = bottom_right
        w = x2 - x1
        h = y2 - y1
        return cls(x1, y1, w, h)

    @classmethod
    def from_origin(cls, origin, size):
        """Alternative constructor using a point and size."""
        x, y = origin
        w, h = size
        return cls(x, y, w, h)

    @classmethod
    def from_center(cls, center, size):
        """Alternative constructor using a center point and size."""
        w, h = size
        xc, yc = center
        x = xc - w/2
        y = yc - h/2
        return cls(x, y, w, h)

    def _slice(self):
        """Returns a slice for a numpy array. Not included in OpenCV.

        img[rect._slice()] == img[rect.y : rect.y + rect.height, rect.x : rect.x + rect.width]
        """
        return slice(self.y, self.y + self.height), slice(self.x, self.x + self.width)

    def _center(self):
        """Returns the center of the rectangle as a point (xc, yc). Not included in OpenCV.

        rect._center() == (rect.x + rect.width / 2, rect.y + rect.height / 2)
        """
        return Point(self.x + self.width / 2, self.y + self.height / 2)


def _floor(v):
    return int(round(np.floor(v)))


def _ceil(v):
    return int(round(np.ceil(v)))


class RotatedRect(NamedTuple):
    center: Point
    size: Size
    angle: float

    def bounding_rect(self):
        """returns the minimal rectangle containing the rotated rectangle"""
        pts = self.points()
        r = Rect(
            Point(_floor(min(pt.x for pt in pts)), _floor(min(pt.y for pt in pts))),
            Point(_ceil(max(pt.x for pt in pts)), _ceil(max(pt.y for pt in pts))),
        )
        return r

    def points(self):
        """returns 4 vertices of the rectangle. The order is bottom left, top left, top right, bottom right."""
        b = np.cos(np.radians(self.angle)) * 0.5
        a = np.sin(np.radians(self.angle)) * 0.5

        pt0 = Point(
            center.x - a * self.size.height - b * self.size.width,
            center.y + b * self.size.height - a * self.size.width,
        )
        pt1 = Point(
            center.x + a * self.size.height - b * self.size.width,
            center.y - b * self.size.height - a * self.size.width,
        )

        pt2 = Point(2 * center.x - pt0.x, 2 * center.y - pt0.y)
        pt3 = Point(2 * center.x - pt1.x, 2 * center.y - pt1.y)

        return [pt0, pt1, pt2, pt3]

    @classmethod
    def from_points(cls, point1, point2, point3):
        """Any 3 end points of the RotatedRect. They must be given in order (either clockwise or anticlockwise)."""
        point1, point2, point3 = Point(*point1), Point(*point2), Point(*point3)
        center = 0.5 * (point1 + point3)
        vecs = [Point(point1 - point2), Point(point2 - point3)]
        x = max(np.linalg.norm(pt) for pt in (point1, point2, point3))
        a = min(np.linalg.norm(vecs[0]), np.linalg.norm(vecs[1]))

        # check that given sides are perpendicular
        if abs(vecs[0].dot(vecs[1])) * a <= np.finfo(np.float32).eps * 9 * x * (
            np.linalg.norm(vecs[0]) * np.linalg.norm(vecs[1])
        ):
            raise ValueError(
                "The three points do not define a rotated rect. The three points should form a right triangle."
            )

        # wd_i stores which vector (0,1) or (1,2) will make the width
        # One of them will definitely have slope within -1 to 1
        wd_i = 1 if abs(vecs[1][1]) < abs(vecs[1][0]) else 0
        ht_i = (wd_i + 1) % 2

        angle = np.degrees(np.atan(vecs[wd_i][1] / vecs[wd_i][0]))
        width = np.linalg.norm(vecs[wd_i])
        height = np.linalg.norm(vecs[ht_i])
        size = Size(width, height)

        return cls(center, size, angle)


class Scalar(tuple):
    def __new__(cls, v0=0, v1=0, v2=0, v3=0):
        return super().__new__(cls, [v0, v1, v2, v3])

    def conj(self):
        v0, v1, v2, v3 = self
        return type(self)(v0, -v1, -v2, -v3)

    def is_real(self):
        v0, v1, v2, v3 = self
        return v1 == v2 == v3 == 0

    def mul(self, other, scale=1):
        other = Scalar(*other)
        return type(self)(scale * v * w for v, w in zip(self, other))

    @classmethod
    def all(cls, v0):
        return cls(v0, v0, v0, v0)


class TermCriteria(NamedTuple):
    type: int = 0
    max_count: int = 0
    epsilon: float = 0

    COUNT = cv.TermCriteria_COUNT
    MAX_ITER = cv.TermCriteria_MAX_ITER
    EPS = cv.TermCriteria_EPS

    def is_valid(self):
        is_count = (self.type & self.COUNT) and self.max_count > 0
        is_eps = (self.type & self.EPS) and not np.isna(self.epsilon)
        return is_count or is_eps



import cv2
import numpy as np


def resize_pad(img, size, pad_color=(0, 0, 0), interpolation=None):

    h, w = img.shape[:2]
    sw, sh = size
    aspect = w / h

    if interpolation is None:
        interpolation = cv2.INTER_AREA if (h > sh) or (w > sw) else cv2.INTER_CUBIC

    # compute scaling and pad sizing
    if aspect > 1:
        nh = int(np.round(sw / aspect))
        new_size = (sw, nh)
        pad_vert = (sh - nh) // 2
        padding = (pad_vert, sh - pad_vert, 0, 0)
    elif aspect < 1:
        nw = int(np.round(sh * aspect))
        new_size = (nw, sh)
        pad_horz = (sw - nw) // 2
        padding = (0, 0, pad_horz, sw - pad_horz)
    else:
        new_size = (sw, sh)
        padding = (0, 0, 0, 0)

    # scale and pad
    scaled_img = cv2.resize(img, new_size, interpolation=interpolation)
    scaled_img = cv2.copyMakeBorder(
        scaled_img, *padding, borderType=cv2.BORDER_CONSTANT, value=pad_color
    )

    return scaled_img


def circular_gradient(w, h, center=None, radius=None, invert=False):
    """Makes a gradient, white at the center point, fading out to entirely black
    at radius. Colors are flipped if invert == True.

    By default, center is the center of (w, h) and radius is the distance to
    the nearer edge.
    """
    center = center or (w // 2, h // 2)
    radius = radius or min(w, h) // 2

    mask = np.zeros((h, w), dtype=np.uint8)
    mask = cv2.circle(mask, center=center, radius=radius, color=255, thickness=-1)
    grad = cv2.distanceTransform(mask, cv2.DIST_L2, 3)
    grad = grad / grad.max()

    if invert:
        grad = 1 - grad

    return grad


def circular_mask(w, h, center=None, radius=None):
    """Creates a circular binary/logical mask.

    Parameters
    ==========
    w, h : int
        Width and height of the mask.
    center : tuple(numeric, numeric) (default: (h/2, w/2))
        The center of the circular mask. 
    radius : tuple(numeric, numeric) (default: nearest image bound to center)
        Radius of the circle extending from the center point. Note that pixels
        that touch the radius will be included. 
    
    Returns
    =======
    mask : np.ndarray
        Boolean array with shape (h, w) where values are True inside the circle,
        False otherwise.

    Notes
    =====
    From this Stack Overflow answer: https://stackoverflow.com/a/44874588/5087436
    """

    if center is None:  # use the middle of the image
        center = [int(w / 2), int(h / 2)]
    if radius is None:  # use the smallest distance between the center and image walls
        radius = min(center[0], center[1], w - center[0], h - center[1])

    Y, X = np.ogrid[:h, :w]
    dist_from_center = np.sqrt((X - center[0]) ** 2 + (Y - center[1]) ** 2)

    return dist_from_center <= radius



"""Padded transformation module.

This module provides two functions, warpPerspectivePadded() and
warpAffinePadded(), which compliment the built-in OpenCV functions
warpPerspective() and warpAffine(). These functions calculate the
extent of the warped image and pads both the destination and the
warped image so both images can be fully displayed together.
"""


import cv2
import numpy as np


def warpPerspectivePadded(
    src, dst, M, flags=cv2.INTER_LINEAR, borderMode=cv2.BORDER_CONSTANT, borderValue=0
):
    """Performs a perspective warp with padding.

    Parameters
    ----------
    src : array_like
        source image, to be warped.
    dst : array_like
        destination image, to be padded.
    M : array_like
        `3x3` perspective transformation matrix.

    Returns
    -------
    src_warped : ndarray
        padded and warped source image
    dst_padded : ndarray
        padded destination image, same size as src_warped

    Optional Parameters
    -------------------
    flags : int, optional
        combination of interpolation methods (`cv2.INTER_LINEAR` or
        `cv2.INTER_NEAREST`) and the optional flag `cv2.WARP_INVERSE_MAP`,
        that sets `M` as the inverse transformation (`dst` --> `src`).
    borderMode : int, optional
        pixel extrapolation method (`cv2.BORDER_CONSTANT` or
        `cv2.BORDER_REPLICATE`).
    borderValue : numeric, optional
        value used in case of a constant border; by default, it equals 0.

    See Also
    --------
    warpAffinePadded : for `2x3` affine transformations
    cv2.warpPerspective : original OpenCV function
    """

    assert M.shape == (3, 3), (
        "Perspective transformation shape should be (3, 3).\n"
        + "Use warpAffinePadded() for (2, 3) affine transformations."
    )

    M = M / M[2, 2]  # ensure a legal homography
    if flags in (
        cv2.WARP_INVERSE_MAP,
        cv2.INTER_LINEAR + cv2.WARP_INVERSE_MAP,
        cv2.INTER_NEAREST + cv2.WARP_INVERSE_MAP,
    ):
        M = cv2.invert(M)[1]
        flags -= cv2.WARP_INVERSE_MAP

    # it is enough to find where the corners of the image go to find
    # the padding bounds; points in clockwise order from origin
    src_h, src_w = src.shape[:2]
    lin_homg_pts = np.array([[0, src_w, src_w, 0], [0, 0, src_h, src_h], [1, 1, 1, 1]])

    # transform points
    transf_lin_homg_pts = M.dot(lin_homg_pts)
    transf_lin_homg_pts /= transf_lin_homg_pts[2, :]

    # find min and max points
    min_x = np.floor(np.min(transf_lin_homg_pts[0])).astype(int)
    min_y = np.floor(np.min(transf_lin_homg_pts[1])).astype(int)
    max_x = np.ceil(np.max(transf_lin_homg_pts[0])).astype(int)
    max_y = np.ceil(np.max(transf_lin_homg_pts[1])).astype(int)

    # add translation to the transformation matrix to shift to positive values
    anchor_x, anchor_y = 0, 0
    transl_transf = np.eye(3, 3)
    if min_x < 0:
        anchor_x = -min_x
        transl_transf[0, 2] += anchor_x
    if min_y < 0:
        anchor_y = -min_y
        transl_transf[1, 2] += anchor_y
    shifted_transf = transl_transf.dot(M)
    shifted_transf /= shifted_transf[2, 2]

    # create padded destination image
    dst_h, dst_w = dst.shape[:2]

    pad_widths = [
        anchor_y,
        max(max_y, dst_h) - dst_h,
        anchor_x,
        max(max_x, dst_w) - dst_w,
    ]

    dst_padded = cv2.copyMakeBorder(
        dst, *pad_widths, borderType=borderMode, value=borderValue
    )

    dst_pad_h, dst_pad_w = dst_padded.shape[:2]
    src_warped = cv2.warpPerspective(
        src,
        shifted_transf,
        (dst_pad_w, dst_pad_h),
        flags=flags,
        borderMode=borderMode,
        borderValue=borderValue,
    )

    return dst_padded, src_warped


def warpAffinePadded(
    src, dst, M, flags=cv2.INTER_LINEAR, borderMode=cv2.BORDER_CONSTANT, borderValue=0
):
    """Performs an affine or Euclidean/rigid warp with padding.

    Parameters
    ----------
    src : array_like
        source image, to be warped.
    dst : array_like
        destination image, to be padded.
    M : array_like
        `2x3` affine transformation matrix.

    Returns
    -------
    src_warped : ndarray
        padded and warped source image
    dst_padded : ndarray
        padded destination image, same size as src_warped

    Optional Parameters
    -------------------
    flags : int, optional
        combination of interpolation methods (`cv2.INTER_LINEAR` or
        `cv2.INTER_NEAREST`) and the optional flag `cv2.WARP_INVERSE_MAP`,
        that sets `M` as the inverse transformation (`dst` --> `src`).
    borderMode : int, optional
        pixel extrapolation method (`cv2.BORDER_CONSTANT` or
        `cv2.BORDER_REPLICATE`).
    borderValue : numeric, optional
        value used in case of a constant border; by default, it equals 0.

    See Also
    --------
    warpPerspectivePadded : for `3x3` perspective transformations
    cv2.warpAffine : original OpenCV function
    """
    assert M.shape == (2, 3), (
        "Affine transformation shape should be (2, 3).\n"
        + "Use warpPerspectivePadded() for (3, 3) homography transformations."
    )

    if flags in (
        cv2.WARP_INVERSE_MAP,
        cv2.INTER_LINEAR + cv2.WARP_INVERSE_MAP,
        cv2.INTER_NEAREST + cv2.WARP_INVERSE_MAP,
    ):
        M = cv2.invertAffineTransform(M)
        flags -= cv2.WARP_INVERSE_MAP

    # it is enough to find where the corners of the image go to find
    # the padding bounds; points in clockwise order from origin
    src_h, src_w = src.shape[:2]
    lin_pts = np.array([[0, src_w, src_w, 0], [0, 0, src_h, src_h]])

    # transform points
    transf_lin_pts = M[:, :2].dot(lin_pts) + M[:, 2].reshape(2, 1)

    # find min and max points
    min_x = np.floor(np.min(transf_lin_pts[0])).astype(int)
    min_y = np.floor(np.min(transf_lin_pts[1])).astype(int)
    max_x = np.ceil(np.max(transf_lin_pts[0])).astype(int)
    max_y = np.ceil(np.max(transf_lin_pts[1])).astype(int)

    # add translation to the transformation matrix to shift to positive values
    anchor_x, anchor_y = 0, 0
    if min_x < 0:
        anchor_x = -min_x
    if min_y < 0:
        anchor_y = -min_y
    shifted_transf = M + [[0, 0, anchor_x], [0, 0, anchor_y]]

    # create padded destination image
    dst_h, dst_w = dst.shape[:2]

    pad_widths = [
        anchor_y,
        max(max_y, dst_h) - dst_h,
        anchor_x,
        max(max_x, dst_w) - dst_w,
    ]

    dst_padded = cv2.copyMakeBorder(
        dst, *pad_widths, borderType=borderMode, value=borderValue
    )

    dst_pad_h, dst_pad_w = dst_padded.shape[:2]
    src_warped = cv2.warpAffine(
        src,
        shifted_transf,
        (dst_pad_w, dst_pad_h),
        flags=flags,
        borderMode=borderMode,
        borderValue=borderValue,
    )

    return dst_padded, src_warped


"""
References
----------
See the following question and my answer on Stack Overflow for an
idea of how this was conceptualized and to read the mathematics
behind the functions: https://stackoverflow.com/a/44459869/5087436

"""



from itertools import tee


def pairwise(iterable):
    """s -> (s0, s1), (s1, s2), ..."""
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


class LineIterator:
    """Iterates through the pixels in a line between two points.

    The line will be clipped on the image boundaries
    The line can be 4- or 8-connected.
    If left_to_right=True, then the iteration is always
    done from the left-most point to the right most,
    not to depend on the ordering of pt1 and pt2 parameters.
    """

    def __init__(self, img, pt1, pt2, connectivity=8, left_to_right=False):

        count = -1
        cols, rows = img.shape[:2]
        self.step = cols
        x1, y1 = pt1
        x2, y2 = pt2

        if not (0, 0, 0, 0) <= (x1, x2, y1, y2) < (cols, cols, rows, rows):
            clip, pt1, pt2 = cv2.clipLine((0, 0, cols, rows), pt1, pt2)
            x1, y1 = pt1
            x2, y2 = pt2
            if not clip:
                self.count = 0
                return

        bt_pix = 1
        istep = self.step

        dx = x2 - x1
        dy = y2 - y1
        s = -1 if dx < 0 else 0

        if left_to_right:
            dx = (dx ^ s) - s
            dy = (dy ^ s) - s
            x1 ^= (x1 ^ x2) & s
            y1 ^= (y1 ^ y2) & s
        else:
            dx = (dx ^ s) - s
            bt_pix = (bt_pix ^ s) - s

        self.index = y1 * istep + x1

        s = -1 if dy < 0 else 0
        dy = (dy ^ s) - s
        istep = (istep ^ s) - s

        # conditional swaps
        s = -1 if dy > dx else 0
        dx ^= dy & s
        dy ^= dx & s
        dx ^= dy & s

        bt_pix ^= istep & s
        istep ^= bt_pix & s
        bt_pix ^= istep & s

        assert dx >= 0 and dy >= 0
        if connectivity == 8:
            self.err = dx - (dy + dy)
            self.plus_delta = dx + dx
            self.plus_step = int(istep)
            self.count = dx + 1
        else:
            self.err = 0
            self.plus_delta = (dx + dx) + (dy + dy)
            self.plus_step = int(istep - bt_pix)
            self.count = dx + dy + 1
        self.minus_delta = -(dy + dy)
        self.minus_step = int(bt_pix)

    def __iter__(self):
        for i in range(self.count):
            y = int(self.index / self.step)
            x = int(self.index - y * self.step)
            yield (x, y)
            mask = -1 if self.err < 0 else 0
            self.err += self.minus_delta + (self.plus_delta & mask)
            self.index += self.minus_step + (self.plus_step & mask)
